#include "B4RDefines.h"
#define Debug 0
namespace B4R {

	//returns a B4R byte array containing date/time representing the seconds given
	ArrayByte* B4RESP8266rtc::Sec2DT(ULong sec)  {
		ArrayByte* datetime = CreateStackMemoryObject(ArrayByte);
		datetime->data = StackMemory::buffer + StackMemory::cp;
		datetime->length = 8;
		B4R::StackMemory::cp += 8;
		
		byte dt[8];
		SecAsDateTime(sec,dt);
		for (byte i = 0;i < 8;i++) {
			((byte*)datetime->data)[i] = dt[i];
		}
		
		return datetime;
	}
	
	/*===================================================================================================*/	
	/* functions to convert to and from system time 										             */
	/* These are for interfacing with time services                                                      */
	/* leap year calulator expects year argument as years offset from 1970					             */
	#define LEAP_YEAR(Y)     ( ((1970+(Y))>0) && !((1970+(Y))%4) && ( ((1970+(Y))%100) || !((1970+(Y))%400) ) )
	// API starts months from 1, this array starts from 0
	static  const uint8_t monthDays[]={31,28,31,30,31,30,31,31,30,31,30,31};
	
	//private function
	//convert seconds into an 8 bytes array containing date/time
	void B4RESP8266rtc::SecAsDateTime(ULong sec, byte* dt)  {
		uint8_t year;
		uint8_t month, monthLength;
		uint32_t time;
		unsigned long days;
		time = (uint32_t)sec;
		int Second = time % 60;
		dt[6] = Second;
		time /= 60; // now it is minutes
		int Minute = time % 60;
		dt[5] = Minute;
		time /= 60; // now it is hours
		int Hour = time % 24;
		dt[4] = Hour;
		time /= 24; // now it is days
		int Wday = ((time + 4) % 7) + 1;  // Sunday is day 1, saturday is 7
		dt[7] = Wday;

		year = 0;  
		days = 0;
		while((unsigned)(days += (LEAP_YEAR(year) ? 366 : 365)) <= time) {
			year++;				// normally for 2019 it's 49 (2019-1970)
		}
		// Calculate the century from 2000 -> work even after 2099, if ever they solve the 2038 bug on servers
		if ((year-30) > 99) {
			int century = (year-30) / 100;
			dt[0] = century;		//Century is the exact number of year(s) passed after 2099
			int year_after_2000 = (year-30) % 100;
			dt[1] = year_after_2000;
		} else {
			dt[0] = 20;
			int year_after_2000 = (year-30);			
			dt[1] = year_after_2000;
		}

		days -= (LEAP_YEAR(year) ? 366 : 365);
		time  -= days;		// now it is days in this year, starting at 0
		days=0;
		month=0;
		monthLength=0;
		for (month=0; month<12; month++) {
			if (month==1) {			// february, array start at 0
				if (LEAP_YEAR(year)) {
					monthLength=29;
				} else {
					monthLength=28;
				}
			} else {
				monthLength = monthDays[month];
			}

			if (time >= monthLength) {
				time -= monthLength;
			} else {
				break;
			}
		}
		int Month = month + 1;				// jan is month 1 
		dt[2] = Month;
		int Day = time + 1;					// day of month
		dt[3] = Day;
	}

	//returns the current date/time as a B4R byte array
	bool B4RESP8266rtc::ReadClock(ArrayByte* datetime)  {
		if (SetAt_millis==0)  {
			return false;  //clock has not been set
		}

		ULong sec;
		sec = (millis() - SetAt_millis)/1000;
		sec = sec + Set_DateTime_AsSeconds;

		byte dt[8];
		SecAsDateTime(sec,dt);
		for (byte i = 0;i < 8;i++) {
			((byte*)datetime->data)[i] = dt[i];
		}
		return true;  //success
	}
	
	//returns a seconds for the B4R byte array containing a date/time
	ULong B4RESP8266rtc::DT2Sec(ArrayByte* datetime)  {
		int i;
		uint32_t seconds;
		int Year = ((byte*)datetime->data)[1];
		// year now will be since 1970 (normal Unix timestamp)
		Year += 30;

		// seconds from 1970 till 1 jan 00:00:00 of the given year
		seconds= Year*(86400 * 365);
		for (i = 0; i < Year; i++) {
			if (LEAP_YEAR(i)) {
				seconds +=  86400;   // add extra days for leap years
			}
		}
  
		// add days for this year, months start from 1
		int Month = ((byte*)datetime->data)[2];
		for (i = 1; i < Month; i++) {
			if ( (i == 2) && LEAP_YEAR(Year)) { 
				seconds += 86400 * 29;
			} else {
				seconds += 86400 * monthDays[i-1];  //monthDay array starts from 0
			}
		}
		int Day = ((byte*)datetime->data)[3];
		seconds+= (Day-1) * 86400;
		int Hour = ((byte*)datetime->data)[4];
		seconds+= Hour * 3600;
		int Minutes = ((byte*)datetime->data)[5];
		seconds+= Minutes * 60;
		int Seconds = ((byte*)datetime->data)[6];
		seconds+= Seconds;
		return seconds; 
	}
	
	void B4RESP8266rtc::SetClock(ArrayByte* datetime, 
							int offset_hours, int offset_minutes)  {

		ULong sec = DT2Sec(datetime);
		ULong offset = ((offset_hours*60)+offset_minutes)*60;
		sec = sec + offset;
		
		Set_DateTime_AsSeconds = sec;
		SetAt_millis = millis();
		
		if (tick_every_second)  {
			start_one_second_tick();
		}
	}
	
	int B4RESP8266rtc::GetUTC(ArrayByte* dt, B4RString* Host, B4RString* Url)  {
		WiFiClient client;
		const int httpPort = 80;

		if (WiFi.status() != WL_CONNECTED) {
			//Wifi not connected
			return 2;
		}
		
		if (!client.connect(Host->data, httpPort)) {
			//Host not found
			return 3;
		}

		client.print(String("GET ") + Url->data + " HTTP/1.1\r\n" +
					"Host: " + Host->data + "\r\n" + 
					"User-Agent: esp8266 \r\n" + 
					"Connection: close\r\n\r\n");
		delay(1000);
		
		while(client.available()) {
			String line = client.readStringUntil('\r');
			if (line=="\n") {
				String line = client.readStringUntil('\r');
				line = line + ' ';
				//convert string to Long
				char buf[line.length()];
				line.toCharArray(buf,line.length());
				ULong TimeInSeconds=atol(buf); 
				//::Serial.println(TimeInSeconds);
				byte xdt[8];
				SecAsDateTime(TimeInSeconds,xdt);
				for (byte i = 0;i < 8;i++) {
					((byte*)dt->data)[i] = xdt[i];
				}
				if (xdt[0]!=0) {
					return 0;   //successfully obtained DateTime
				} else {
					return 1;   //invalid Date/Time
				}
			}
		}

		
		while(client.available()) {
			String line = client.readStringUntil('\r');
			client.stop();
			//convert String to Long
			char buf[line.length()];
			::Serial.println(line);
			line.toCharArray(buf,line.length());
			long TimeInSeconds = atol(buf); 
			//Serial.println(result);
			return TimeInSeconds;
			}
		return 4;  //invalid response from host
	}

	long string_to_long (String number)
	{
	  number = number + ' ';
	  char buf[number.length()];
	  //Serial.println(number);
	  number.toCharArray(buf,number.length());
	  long result=atol(buf); 
	  //Serial.println(result);
	  return result;
	}

	// Get 8 byte Date/Time array from a server
	int B4RESP8266rtc::GetDT(ArrayByte* dt, B4RString* Host, UInt Port)  {
		byte xdt[8];
		WiFiClient client;

		if (WiFi.status() != WL_CONNECTED) {
			//Wifi not connected
			return 2;
		}
		
		if (!client.connect(Host->data, Port)) {
			//Host not found
			return 3;
		}

		delay(1000);
		
		while(client.available()) {
			int b;
			for (byte i = 0;i < 8;i++) {
				b = client.read();
				if (b == -1)  {
					client.stop();
					return 4;   //invalid response from server
				}
				((byte*)dt->data)[i] = b;
			}
			client.stop();
			if ((xdt[0]!=0) && (xdt[7]==0)) {
				return 0;   //successfully obtained DateTime
			} else {
				return 1;   //invalid Date/Time
			}
		}
		return 4;  //invalid response from host
	}

#include <WiFiUdp.h>
#define NTP_PACKET_SIZE 48
	// Get 8 byte Date/Time array from a server
	int B4RESP8266rtc::GetNTP(ArrayByte* dt, B4RString* Host, UInt Port)  {
		unsigned int localPort = 2390;      // local port to listen for UDP packets
		byte packetBuffer[NTP_PACKET_SIZE]; //buffer to hold incoming and outgoing packets

		if (WiFi.status() != WL_CONNECTED) {
			//Wifi not connected
			return 2;
		}
		
		WiFiUDP udp;
		udp.begin(localPort);

		memset(packetBuffer, 0, NTP_PACKET_SIZE);
		// Initialize values needed to form NTP request
		packetBuffer[0] = 0b11100011;   // LI, Version, Mode
		packetBuffer[1] = 0;     // Stratum, or type of clock
		packetBuffer[2] = 6;     // Polling Interval
		packetBuffer[3] = 0xEC;  // Peer Clock Precision
		// 8 bytes of zero for Root Delay & Root Dispersion
		packetBuffer[12]  = 49;
		packetBuffer[13]  = 0x4E;
		packetBuffer[14]  = 49;
		packetBuffer[15]  = 52;

		// all NTP fields have been given values, now
		// you can send a packet requesting a timestamp:
		udp.beginPacket(Host->data, Port); 		//NTP requests are normally on port 123
		udp.write(packetBuffer, NTP_PACKET_SIZE);
		udp.endPacket();
		// wait to see if a reply is available
		delay(2000);

		bool cb = udp.parsePacket();
		if (!cb) {
			udp.stop();
			return 3;   //Invalid response from server
		} else {
			// We've received a packet, read the data from it
			udp.read(packetBuffer, NTP_PACKET_SIZE); // read the packet into the buffer
			//the timestamp starts at byte 40 of the received packet and is four bytes,
			// or two words, long. First, esxtract the two words:
			unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
			unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
			// combine the four bytes (two words) into a long integer
			// this is NTP time (seconds since Jan 1 1900):
			unsigned long secsSince1900 = highWord << 16 | lowWord;
			// Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
			const unsigned long seventyYears = 2208988800UL;
			// subtract seventy years:
			unsigned long epoch = secsSince1900 - seventyYears;
#if Debug
			::Serial.print("Epoch: ");
			::Serial.println(epoch);
#endif
			udp.stop();
			// convert to the array of byte
			byte xdt[8];
			SecAsDateTime(epoch, xdt);
#if Debug
			::Serial.print("xdt: ");
			::Serial.print(xdt[0]);
			::Serial.print(", ");
			::Serial.print(xdt[1]);
			::Serial.print(", ");
			::Serial.print(xdt[2]);
			::Serial.print(", ");
			::Serial.print(xdt[3]);
			::Serial.print(", ");
			::Serial.print(xdt[4]);
			::Serial.print(", ");
			::Serial.print(xdt[5]);
			::Serial.print(", ");
			::Serial.print(xdt[6]);
			::Serial.print(", ");
			::Serial.println(xdt[7]);
#endif
			for (byte i = 0;i < 8;i++) {
				((byte*)dt->data)[i] = xdt[i];
			}
#if Debug
			::Serial.print("dt: ");
			::Serial.print(((byte*)dt->data)[0]);
			::Serial.print(", ");
			::Serial.print(((byte*)dt->data)[1]);
			::Serial.print(", ");
			::Serial.print(((byte*)dt->data)[2]);
			::Serial.print(", ");
			::Serial.print(((byte*)dt->data)[3]);
			::Serial.print(", ");
			::Serial.print(((byte*)dt->data)[4]);
			::Serial.print(", ");
			::Serial.print(((byte*)dt->data)[5]);
			::Serial.print(", ");
			::Serial.print(((byte*)dt->data)[6]);
			::Serial.print(", ");
			::Serial.println(((byte*)dt->data)[7]);
#endif
			if (xdt[0]!=0) {
				return 0;   //successfully obtained DateTime
			} else {
				return 1;   //invalid Date/Time
			}
		}
	}

	bool B4RESP8266rtc::isNowDST(UInt forWhere, ArrayByte* datetime, int tzHours){
		if (forWhere > 1) {
			return false;
		}
		int year = ((byte*)datetime->data)[0] * 100;
		year += ((byte*)datetime->data)[1];
#if Debug
			::Serial.print("year: ");
			::Serial.println(year);
#endif
		byte month = ((byte*)datetime->data)[2];
		byte day = ((byte*)datetime->data)[3];
		byte hour = ((byte*)datetime->data)[4];
		if (forWhere == 0) {		// North America DST state, DST start 2nd sunday of march, till 1st Sunday of November
#if Debug
			::Serial.print("value with hour to check: ");
			::Serial.println((hour + 24 * day));
			::Serial.print("DST begin to check: ");
			::Serial.println((1 + tzHours + 24*(14 - (1 + year * 5/4) % 7)));
			::Serial.print("DST end to check: ");
			::Serial.println((1 + tzHours + 24*(7 - (1 + year * 5/4) % 7)));
#endif
			if (month<3 || month>11) return false;	// it's not DST
			if (month>3 && month<11) return true;   // it's DST
			if (month==3 && (hour + 24 * day) >= (1 + tzHours + 24*(14 - (1 + year * 5/4) % 7))
				|| month==11 && (hour + 24 * day) < (1 + tzHours + 24*(7 - (1 + year * 5/4) % 7))) {
				return true;	// it's DST
			} else {
				return false;	// it's not DST
			}
		} 
		if (forWhere == 1) {		// Europe DST state, DST sart last sunday of march, till last sunday of october
#if Debug
			::Serial.print("value with hour to check: ");
			::Serial.println((hour + 24 * day));
			::Serial.print("DST begin to check: ");
			::Serial.println((1 + tzHours + 24*(31 - (5 * year/4 + 4) % 7)));
			::Serial.print("DST end to check: ");
			::Serial.println((1 + tzHours + 24*(31 - (5 * year/4 + 1) % 7)));
#endif
			if (month<3 || month>10) return false;	// it's not DST
			if (month>3 && month<10) return true;   // it's DST
			if (month==3 && (hour + 24 * day) >= (1 + tzHours + 24*(31 - (5 * year/4 + 4) % 7))
				|| month==10 && (hour + 24 * day) < (1 + tzHours + 24*(31 - (5 * year/4 + 1) % 7))) {
				return true;	// it's DST
			} else {
				return false;	// it's not DST
			}
		}
	}  

	static union FunctionUnion fncu;
	void B4RESP8266rtc::Initialize(void (*Event) (byte idx), bool tick_every_sec) {
		fncu.TimerFunction = tick;
		this->event = Event;
	
		for (byte i = 0;i < 10;i++) {
			this->enabled[i] = 0;  //0=not active
		}
		
		tick_every_second = tick_every_sec;
	}

	void B4RESP8266rtc::start_one_second_tick() {
		counter++;
		if (counter==0)  counter=1;
		this->enabled[0] = counter;
		scheduler.add(millis() + 1000, fncu, counter, this);
	}
	
	bool B4RESP8266rtc::SetAlarm(byte Index, ArrayByte* datetime) {
		if ((Index==0) || (Index>9))  {
			return false;   //index out of range (1 to 9 only)
		}
		
		ULong CurrentAsSec;
		CurrentAsSec = (millis() - SetAt_millis)/1000;
		CurrentAsSec = CurrentAsSec + Set_DateTime_AsSeconds;
		ULong AlmAsSec = DT2Sec(datetime);
		if (AlmAsSec<=CurrentAsSec)  {
			return false;   //date/time has already passed
		}
		ULong Interval = (AlmAsSec - CurrentAsSec);
		Interval = Interval * 1000;
		
		counter++;
		if (counter==0)  counter=1;
		this->enabled[Index] = counter;
		scheduler.add(millis() + Interval, fncu, counter, this);
		return true;  //success
	}
	bool B4RESP8266rtc::CancelAlarm(byte Index)  {
		if ((Index==0) || (Index>9))  {
			return false;   //index out of range (0 to 9 only)
		}
		this->enabled[Index] = 0;  //0=not active
		return true;  //success
	}
	
	//called on alarm event
	void B4RESP8266rtc::tick(byte tag, void* target) {
		B4RESP8266rtc* me = (B4RESP8266rtc*)target;

		for (byte i = 0;i < 10;i++) {
			if (me->enabled[i] == tag)  {
				me->enabled[i] = 0;  //0=not active
				if (i==0)  {
					me->start_one_second_tick();  //and the next second, and so on
				}
				me->event(i);  //raise B4R alarm event with it's "Index" (0 to 9)
			}
		}
	}

	B4RString* B4RESP8266rtc::ShortDate(ArrayByte* datetime) {
		PrintToMemory pm;
		B4RString* s = B4RString::PrintableToString(NULL);
		
		byte v;
		v =((byte*)datetime->data)[3];  //day
		if (v<10) pm.print(0);
		pm.print(v);
		
		StackMemory::buffer[StackMemory::cp++] = (byte)'/';		
		
		v =((byte*)datetime->data)[2];  //month
		v=((v-1)*3);
		String months = "JanFebMarAprMayJunJulAugSepOctNovDec";
		for (byte i = 0;i < 3;i++) {
			StackMemory::buffer[StackMemory::cp++] = (byte)months[v+i];		
		}
		
		StackMemory::buffer[StackMemory::cp++] = (byte)'/';	
		
		v =((byte*)datetime->data)[1];  //year
		pm.print(v);
		StackMemory::buffer[StackMemory::cp++] = 0;  //end of string
		
		return s;  //short date as string
	}

		B4RString* B4RESP8266rtc::Time24(ArrayByte* datetime) {
		PrintToMemory pm;
		B4RString* s = B4RString::PrintableToString(NULL);
		
		byte v;
		v =((byte*)datetime->data)[4];  //hours
		if (v<10) pm.print(0);
		pm.print(v);
		
		StackMemory::buffer[StackMemory::cp++] = (byte)':';		
		
		v =((byte*)datetime->data)[5];  //minutes
		if (v<10) pm.print(0);
		pm.print(v);
		
		StackMemory::buffer[StackMemory::cp++] = (byte)':';		
		
		v =((byte*)datetime->data)[6];  //seconds
		if (v<10) pm.print(0);
		pm.print(v);

		StackMemory::buffer[StackMemory::cp++] = 0;  //end of string
		
		return s;  //24hr time
	}

		B4RString* B4RESP8266rtc::Time12(ArrayByte* datetime) {
		PrintToMemory pm;
		B4RString* s = B4RString::PrintableToString(NULL);
		
		byte h;
		h =((byte*)datetime->data)[4];  //hours
		byte v=h%12;
		if (v==0) v=12;
		if (v<10) pm.print(0);
		pm.print(v);

		StackMemory::buffer[StackMemory::cp++] = (byte)':';		
		
		v =((byte*)datetime->data)[5];  //minutes
		if (v<10) pm.print(0);
		pm.print(v);
		
		if (h<12) {
			StackMemory::buffer[StackMemory::cp++] = (byte)'a';			
		} else {
			StackMemory::buffer[StackMemory::cp++] = (byte)'p';			
		}
		StackMemory::buffer[StackMemory::cp++] = (byte)'m';		
		
		StackMemory::buffer[StackMemory::cp++] = 0;  //end of string
		
		return s;  //12hr time
	}

	B4RString* B4RESP8266rtc::ShortDOW(ArrayByte* datetime) {
		//day of week
		PrintToMemory pm;
		B4RString* s = B4RString::PrintableToString(NULL);
		
		byte v;
		v =((byte*)datetime->data)[7];  //day of week
		v=(v*3);
		String dow = "   SunMonTueWedThuFriSat";
		for (byte i = 0;i < 3;i++) {
			StackMemory::buffer[StackMemory::cp++] = (byte)dow[v+i];		
		}
		StackMemory::buffer[StackMemory::cp++] = 0;  //end of string
		
		return s;  //short day-of-week as string
	}

	
	//returns a B4R byte array containing date/time representing the seconds given
	ArrayByte* B4RESP8266rtc::ThisTime(byte this_hour, byte this_minute, byte this_second)  {
		ULong CurrentAsSec;
		CurrentAsSec = (millis() - SetAt_millis)/1000;
		CurrentAsSec = CurrentAsSec + Set_DateTime_AsSeconds;
		
		byte dt[8];
		SecAsDateTime(CurrentAsSec,dt);
		
		byte current_hour = dt[4];
		byte current_minute = dt[5];
		byte current_second = dt[6];
		
		ulong current_time = (current_hour*60*60) + (current_minute*60) + current_second;
		ulong this_time = (this_hour*60*60) + (this_minute*60) + this_second;
		
		if (this_time <= current_time) {
			//time is after mid-night
			ULong TomorrowAsSec = CurrentAsSec + (24*60*60); 
			SecAsDateTime(TomorrowAsSec,dt);
		}

		dt[4] = this_hour;
		dt[5] = this_minute;
		dt[6] = this_second;
		
		ArrayByte* datetime = CreateStackMemoryObject(ArrayByte);
		datetime->data = StackMemory::buffer + StackMemory::cp;
		datetime->length = 8;
		B4R::StackMemory::cp += 8;
		
		for (byte i = 0;i < 8;i++) {
			((byte*)datetime->data)[i] = dt[i];
		}
		
		return datetime;
	}	
	
}
