#pragma once
#include "B4RDefines.h"

#ifdef ESP32
  #include <WiFi.h>
  #include <WiFiClientSecure.h>
#else
  #include <ESP8266WiFi.h>
#endif

//~version: 1.06

namespace B4R {
	//~shortname: ESP8266rtc
	class B4RESP8266rtc {
		public:
			int GetDT(ArrayByte* dt, B4RString* Host, UInt Port);
			int GetUTC(ArrayByte* dt, B4RString* Host, B4RString* Url); 	
			void SetClock(ArrayByte* datetime, int offset_hours, int offset_minutes);			
			bool ReadClock(ArrayByte* datetime);
			void Initialize(void (*Event) (byte idx), bool tick_every_sec);
			bool SetAlarm(byte Index, ArrayByte* datetime);
			bool CancelAlarm(byte Index);
			int GetNTP(ArrayByte* dt, B4RString* Host, UInt Port);
			bool isNowDST(UInt forWhere, ArrayByte* datetime, int tzHours);
			
			ULong DT2Sec(ArrayByte* datetime);
			ArrayByte* Sec2DT(ULong sec);

			static const ULong SECONDS_1MINUTE   = 60;	
			static const ULong SECONDS_1HOUR   = 60*60;
			static const ULong SECONDS_1DAY   = 60*60*24;
			static const ULong SECONDS_1WEEK   = 60*60*24*7;
			
			B4RString* ShortDate(ArrayByte* datetime);
			B4RString* Time12(ArrayByte* datetime);
			B4RString* Time24(ArrayByte* datetime);
			B4RString* ShortDOW(ArrayByte* datetime);
			ArrayByte* ThisTime(byte this_hour, byte this_minute, byte this_second);
		
		private:
			void start_one_second_tick();
			void SecAsDateTime(ULong sec, byte* dt);
			ULong SetAt_millis = 0;
			ULong Set_DateTime_AsSeconds = 0;
			bool tick_every_second = false;
			byte counter = 0;
			byte enabled[10];
			void (*event)(byte idx);
			static void tick(Byte tag, void* target);
	};
}
